/**********************************
RSS and Atom feed parsing script
Version 1.0
Written by Tarquin Wilton-Jones
Copyright 2018 Vivaldi Technologies
**********************************/

/*eslint-disable no-undef*/

//This library parses RSS and Atom feeds, as well as OPML documents.
//It is not designed to validate feeds, and intentionally copes with a wide range of invalid formats, and tries to extract meaning from them.
//For example, it will accept Atom tags inside RSS feeds, and Atom type attributes on RSS title elements.
//It allows broken HTML to be embedded inside feeds, and tries to recover from them.
//It allows HTML in various fields (denoted by property names), and automatically detects the use of HTML strings in RSS feeds, or XHTML in Atom feeds.
//It allows HTML in all versions of RSS, not just 0.92 and 2.0, since authors make mistakes and use it with RSS 0.91 too
//It allows folder structures in OPML using nested outlines
//Make sure retrieval (eg. XMLHttpRequest) forces interpretation as XML, overriding the mimetype

//let feedParser = new FeedParser();
//let parsed = feedParser.parseFeed(xhr.responseXML, xhr.responseText, url);
//parseFeed(newsfeed XML DOM, string to be parsed using the HTML parser if the XML DOM is invalid, base href);
//when an XML string (eg. XML retrieved from a database) needs to be parsed:
//let parsed = feedParser.parseFeed(
//  (new DOMParser()).parseFromString(xmlString, 'application/xml'),
//  xmlString,
//  baseUrl
//);
//let parsed = feedParser.parseOPML(xhr.responseXML, url);
//parseOPML(OPML XML DOM, base href);

// https://validator.w3.org/feed/docs/atom.html
// https://validator.w3.org/feed/docs/rfc4287.html
// http://www.xml.com/pub/a/2004/06/16/dive.html
// http://www.rssboard.org/rss-0-9-0
// http://www.rssboard.org/rss-0-9-1 or http://www.rssboard.org/rss-0-9-1-netscape
// http://www.rssboard.org/rss-0-9-2
// http://www.rssboard.org/rss-specification
// https://validator.w3.org/feed/docs/rss2.html
// https://www.w3schools.com/xml/xml_rss.asp
// http://web.resource.org/rss/1.0/modules/syndication/
// http://dev.opml.org/
// http://www.therssweblog.com/?guid=20051003145153

//the main class used for parsing feeds

function FeedParser () {}

//classes used to store the various feed data types

function NewsFeed () {
	this.category = [];            //RSS and Atom
	this.cloud = [];               //RSS
	this.people = new NewsPeople();//RSS and Atom
	this.feeditems = [];           //RSS and Atom
	this.generator = null;         //RSS and Atom
	this.haserrors = false;        //Combined
	this.htmlrights = '';          //HTML: RSS and Atom
	this.htmldescription = '';     //HTML: RSS and Atom
	this.htmltitle = '';           //HTML: RSS and Atom
	this.icon = null;              //Atom
	this.id = '';                  //Atom
	this.image = null;             //RSS and Atom
	this.language = '';            //RSS and Atom
	this.lastbuilddate = '';       //RSS
	this.links = new NewsLinks();  //RSS and Atom
	this.logo = null;              //Atom
	this.plaintitle = '';          //Combined
	this.pubdate = '';             //RSS
	this.publishdate = '';         //Combined
	this.rating = '';              //RSS
	this.skipdays = new NewsDay(); //RSS
	this.skiphours = [];           //RSS
	this.textinput = [];           //RSS
	this.ttl = '';                 //RSS and Syndication (allowed here in Atom too)
	this.updated = '';             //Atom
	this.updatedate = '';          //Combined
	this.updatefrequency = '';     //Syndication (RSS, but allowed here in Atom too)
	this.updateperiod = '';        //Syndication (RSS, but allowed here in Atom too)
}

function NewsItem () {
	this.category = [];            //RSS and Atom
	this.people = new NewsPeople();//RSS and Atom
	this.htmlbody = '';            //HTML: Combined
	this.htmlcontent = '';         //HTML: Atom
	this.htmldescription = '';     //HTML: RSS
	this.htmlrights = '';          //HTML: Atom
	this.htmlsummary = '';         //HTML: Atom
	this.htmltitle = '';           //HTML: RSS and Atom
	this.id = '';                  //RSS and Atom
	this.links = new NewsLinks();  //RSS and Atom
	this.plaintitle = '';          //RSS and Atom
	this.pubdate = '';             //RSS
	this.publishdate = '';         //Combined
	this.published = '';           //Atom
	this.source = null;            //Atom
	this.updated = '';             //Atom
	this.updatedate = '';          //Combined
	this.xmloriginal = '';         //Combined
}

function NewsImage () {
	//RSS's elements are supposed to be used to generate this:
	//<a href="link" title="description"><img src="url" alt="title" width="width" height="height"></a>
	this.description = '';         //RSS
	this.height = '';              //RSS
	this.link = '';                //RSS
	this.title = '';               //RSS
	this.url = '';                 //RSS and Atom
	this.width = '';               //RSS
}

function NewsLinks () {
	this.alternate = [];           //RSS and Atom, main link to the article, theoretically there should only be one entry per item+mimetype+hreflang combination
	this.comments = [];            //RSS, link to article comments (item level)
	this.docs = [];                //RSS, not useful, documentation of the feed's RSS format (feed level)
	this.enclosure = [];           //RSS and Atom, usually media files, eg. podcasts (item level)
	this.next = [];                //Atom, not useful, view next feed archive (feed level)
	this.prev = [];                //Atom, not useful, view previous feed archive (feed level)
	this.related = [];             //Atom, other "related articles" links
	this.self = [];                //Atom, not useful, self-referencing link to the feed (feed level)
	this.serviceedit = [];         //Atom, not useful, link to "edit this post" (item level)
	this.servicepost = [];         //Atom, not useful, link to "create new blog post" (feed level)
	this.via = [];                 //RSS and Atom, cited sources (item level)
}

function NewsLink () {
	this.hreflang = '';            //Atom
	this.bytelength = '';          //RSS and Atom
	this.mimetype = '';            //RSS and Atom
	this.title = '';               //Atom
	this.url = '';                 //RSS and Atom
}

function NewsCategory () {
	this.label = '';               //Atom
	this.scheme = '';              //Atom
	this.term = '';                //Atom
}

function NewsPeople () {
	this.author = [];              //Atom for feed, RSS and Atom for items
	this.contributor = [];         //Atom
	this.managingeditor = [];      //RSS (feed level)
	this.webmaster = [];           //RSS (feed level)
}

function NewsPerson () {
	this.email = '';               //RSS and Atom
	this.name = '';                //Atom
	this.url = '';                 //Atom
}

function NewsUA () {
	this.name = '';                //Atom
	this.url = '';                 //Atom
	this.version = '';             //Atom
}

function NewsCloud () {
	//allows the reader to register with a service to be notified of feed updates - unlikely to be used
	this.domain = '';              //RSS
	this.path = '';                //RSS
	this.port = '';                //RSS
	this.protocol = '';            //RSS
	this.registerprocedure = '';   //RSS
}

function NewsInput () {
	//allows the reader to show a form which might be used for feedback - never used
	this.htmldescription = '';     //HTML: RSS
	this.link = '';                //RSS
	this.name = '';                //RSS
	this.title = '';               //RSS
}

function NewsDay () {
	this.sunday =                  //RSS
	this.monday =                  //RSS
	this.tuesday =                 //RSS
	this.wednesday =               //RSS
	this.thursday =                //RSS
	this.friday =                  //RSS
	this.saturday = false;         //RSS
}

function NewsList () {
	this.htmltitle = '';           //OPML
	this.items = [];               //OPML
	this.ownername = '';           //OPML
	this.owneremail = '';          //OPML
	this.ownerurl = '';            //OPML
	this.plaintitle = '';          //OPML
	this.publishdate = '';         //OPML
	this.updatedate = '';          //OPML
}

function NewsOutline () {
	this.category = '';            //OPML
	this.created = '';             //OPML
	this.feedurl = '';             //OPML
	this.htmldescription = '';     //OPML
	this.htmltitle = '';           //OPML
	this.items = [];               //OPML
	this.language = '';            //OPML
	this.pageurl = '';             //OPML
	this.plaintitle = '';          //OPML
	this.supported = true;         //OPML
}

//common namespaces

const namespaceXHTML = 'http://www.w3.org/1999/xhtml';

//common functions

//string manipulation

function trim (str) {
	//trim all leading and trailing whitespace
	return (str || '').replace(/^\s+|\s+$/g, '');
}

function trimLines (str) {
	//trim all leading and trailing whitespace lines, but not leading tabs or spaces before text - for use with plaintext that can be PREformatted
	return (str || '').replace(/^\s+[\r\n]|\s+$/g, '');
}

function htmlSpecialChars (plainText) {
	//encode all important HTML characters in &xx; entities form
	return document.createElementNS(namespaceXHTML, 'div').appendChild(document.createTextNode(plainText)).parentNode.innerHTML;
}

function getPlainString (htmlString) {
	//convert from HTML to plain text
	const divElement = document.createElementNS(namespaceXHTML, 'div');
	divElement.innerHTML = htmlString;
	//if the htmlString started life as plain text, it would have been treated with trimLines only, so can have leading whitespace - that can be removed
	return trim(divElement.textContent);
}

function getAutoHTMLString (htmlString) {
	if (htmlString.match(/&(#\d+|#x[0-9a-e]+|[a-z][0-9a-z]*);|<[a-z][a-z0-9\-]*\b[^>]*>/i)) {
		//HTML is detected by the presence of &foo; &#1234; &#x12ab; <tag...>
		return trim(htmlString);
	} else {
		//it doesn't look like HTML, so encode any special characters as entities
		return htmlSpecialChars(trimLines(htmlString));
	}
}

//DOM string normalizing

function simplifiedName (localName) {
	//just the name of the element or attribute, no namespace, and all in lower case, in case the feed author used CAPS or TitleCase
	return localName.toLowerCase();
}

function simplifiedAttr (attrValue, defaultValue) {
	//lower case attribute name
	return trim((attrValue || defaultValue)).toLowerCase();
}

//DOM manipulation

function concatenateText (elementNode) {
	//concatenate multiple text nodes and CDATA nodes into one
	let plainString = '';
	for (let childIndex = 0; childIndex < elementNode.childNodes.length; childIndex++) {
		if (elementNode.childNodes[childIndex].nodeType === Node.TEXT_NODE || elementNode.childNodes[childIndex].nodeType === Node.CDATA_SECTION_NODE) {
			plainString += elementNode.childNodes[childIndex].nodeValue;
		}
	}
	return plainString;
}

function getAttributeNSs (elementNode, namespaces, attrName) {
	//get a named attribute with one of the namespaces - prefer the last one, since that is the standard namespace when there are multiples
	//prefixed attributes have an explicit namespace
	//by default, unprefixed attributes have no namespace, but inherit a virtual implicit namespace from their element
	//these are the same attribute as each other if the prefix matches the element's namespace
	for (let namespaceIndex = namespaces.length - 1; namespaceIndex >= 0; namespaceIndex--) {
		if (namespaces[namespaceIndex]) {
			//non-empty namespace
			if (elementNode.hasAttributeNS(namespaces[namespaceIndex], attrName)) {
				//explicit namespace, no matter what namespace the element has
				return elementNode.getAttributeNS(namespaces[namespaceIndex], attrName);
			} else if (elementNode.namespaceURI === namespaces[namespaceIndex] && (!namespaceIndex || elementNode.hasAttributeNS('', attrName))) {
				//get an implicitly named attribute, if there is one, or return null on the last loop if not
				return elementNode.getAttributeNS('', attrName);
			} else if (!namespaceIndex) {
				//last loop, there is neither an explicit nor implicit attribute
				return null;
			}
		} else {
			//empty namespace
			if (!namespaceIndex && elementNode.namespaceURI) {
				//in this case, all attributes without a prefix will appear to have no namespace (according to getAttributeNS)
				//but they inherit the implicit namespace behaviour from the element, so return null if this is the last loop
				return null;
			} else if (!namespaceIndex || (!elementNode.namespaceURI && elementNode.hasAttributeNS('', attrName))) {
				//the element has no namespace so any attributes without an explicit namespace also have no implicit namespace
				return elementNode.getAttributeNS('', attrName);
			}
		}
	}
}

function hasAttributeNSs (elementNode, namespaces, attrName) {
	//search for a named attribute with one of the namespaces - prefer the last one, since that is the standard namespace when there are multiples
	//prefixed attributes have an explicit namespace
	//by default, unprefixed attributes have no namespace, but inherit a virtual implicit namespace from their element
	//these are the same attribute as each other if the prefix matches the element's namespace
	for (let namespaceIndex = namespaces.length - 1; namespaceIndex >= 0; namespaceIndex--) {
		if (namespaces[namespaceIndex]) {
			//non-empty namespace
			if (elementNode.hasAttributeNS(namespaces[namespaceIndex], attrName)) {
				//explicit namespace, no matter what namespace the element has
				return true;
			} else if (elementNode.namespaceURI === namespaces[namespaceIndex] && (!namespaceIndex || elementNode.hasAttributeNS('', attrName))) {
				//an implicitly named attribute, if there is one, or return null on the last loop if not
				//if execution reaches here when namespaceIndex is !0, then hasAttributeNS already returned true, avoid a second call to it
				return !!namespaceIndex || elementNode.hasAttributeNS('', attrName);
			} else if (!namespaceIndex) {
				//last loop, there is neither an explicit nor implicit attribute
				return false;
			}
		} else {
			//empty namespace
			if (!namespaceIndex && elementNode.namespaceURI) {
				//in this case, all attributes without a prefix will appear to have no namespace (according to hasAttributeNS)
				//but they inherit the implicit namespace behaviour from the element, so return null if this is the last loop
				return false;
			} else if (!namespaceIndex || (!elementNode.namespaceURI && elementNode.hasAttributeNS('', attrName))) {
				//the element has no namespace so any attributes without an explicit namespace also have no implicit namespace
				//if execution reaches here when namespaceIndex is !0, then hasAttributeNS already returned true, avoid a second call to it
				return !!namespaceIndex || elementNode.getAttributeNS('', attrName);
			}
		}
	}
}

function resolveURL (url, elementNode, currentURL) {
	const scheme = /^[a-z][a-z\-\+\.]+:/i;
	if (url.match(scheme)) {
		//simple case, absolute URL with a scheme/protocol
		return url;
	}
	const baseHref = [url];
	let oneStep, hasAbsolute = false;
	//base:href allows relative bases that refer to each other, so each little step needs to be resolved
	//step up all parents until reaching the documentElement or reaching an element with a non-empty xml:base
	//when doing this inside a reparsed-as-HTML feed, there will be empty body and html, but these will not have any base:href attribute, so it is harmless
	while (!hasAbsolute && elementNode.nodeType === Node.ELEMENT_NODE) {
		if (elementNode.hasAttribute('xml:base')) {
			oneStep = trim(elementNode.getAttribute('xml:base'));
			if (oneStep.length) {
				//stop stepping up if a URL has a scheme
				hasAbsolute = oneStep.match(scheme);
				baseHref[baseHref.length] = oneStep;
			}
		}
		elementNode = elementNode.parentNode;
	}
	if (!hasAbsolute) {
		//no absolute URLs were found, resort to using the feed URL
		baseHref[baseHref.length] = currentURL;
	}
	currentURL = baseHref.pop();
	//xml:base does not affect the .href property so it cannot be used to resolve URLs
	//<base href> applies to a whole document, so resolving must happen in a separate document, not an element or fragment belonging to the current document
	const newDocument = document.implementation.createDocument('http://www.w3.org/1999/xhtml', 'html', null);
	const base = newDocument.documentElement.appendChild(newDocument.createElement('base'));
	//does not actually need to be appended, since it already belongs to that document, but it feels better to do it anyway
	const link = newDocument.documentElement.appendChild(newDocument.createElement('a'));
	//parenthesis and undefined test make sure ESLint does not complain about this valid syntax
	while ((oneStep = baseHref.pop()) !== window.undefined) {
		//there is an xml:base, resolve the link
		base.href = currentURL;
		link.href = oneStep;
		currentURL = link.href;
	}
	return currentURL;
}

//the main feed parsing method

FeedParser.prototype.parseFeed = function (documentNode, htmlString, feedURL) {

	//namespaces

	const namespaceAtom = 'http://www.w3.org/2005/Atom';
	const namespaceFOAF = 'http://xmlns.com/foaf/0.1';
	const namespaceRDF = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#';
	const namespaceSyndication = 'http://purl.org/rss/1.0/modules/syndication/';
	const namespaceRSS = 'http://purl.org/rss/1.0/';
	const namespaceRSS09 = 'http://channel.netscape.com/rdf/simple/0.9/';

	//DOM manipulation

	function stringifyHTML (parentNode, preserveNamespace) {
		let innerXML = '';
		const xmlSerializer = new XMLSerializer();
		parentNode = parentNode.cloneNode(true);
		//this should be compatible with HTML, so CDATA nodes need to be replaced with textNodes
		const treeWalker = document.createTreeWalker(parentNode, NodeFilter.SHOW_CDATA_SECTION);
		while (treeWalker.nextNode()) {
			treeWalker.currentNode.parentNode.replaceChild(treeWalker.currentNode.ownerDocument.createTextNode(treeWalker.currentNode.nodeValue), treeWalker.currentNode);
		}
		if (preserveNamespace) {
			//concatenating children causes it to copy any namespace attribute onto each child element
			for (let childIndex = 0; childIndex < parentNode.childNodes.length; childIndex++) {
				innerXML += xmlSerializer.serializeToString(parentNode.childNodes[childIndex]);
			}
		} else {
			//copy any explicit namespace attributes on child nodes, but remove the parent node which would force namespace inheritance
			innerXML = xmlSerializer.serializeToString(parentNode).replace(/^<[^>]+>|<[^>]+>$/g, '');
		}
		return innerXML;
	}

	function hasNonEmptyTextChild (elementNode) {
		//check whether there are any child text nodes that are not just whitespace, or any CDATA nodes
		for (let childIndex = 0; childIndex < elementNode.childNodes.length; childIndex++) {
			if ((elementNode.childNodes[childIndex].nodeType === Node.TEXT_NODE && elementNode.childNodes[childIndex].nodeValue.replace(/^\s+|\s+$/g, '')) || elementNode.childNodes[childIndex].nodeType === Node.CDATA_SECTION_NODE) {
				return true;
			}
		}
		return false;
	}

	function getElementsByTagNameNSs (parentElement, namespaces, localName) {
		//get an ordered list of all elements in all namespaces with this tag name
		if (namespaces.length === 1) {
			//simple case, just return the native result
			return parentElement.getElementsByTagNameNS(namespaces[0], localName);
		}
		//the following code is used when there is an invalid Atom feed with missing namespace on the root node, and
		//allows correctly namespaced and non-namespaced <entry> elements to be used in the same feed (if the author
		//of the feed is blindly copy & pasting entries with/without namespaces
		//it also allows RSS feeds using both the RDF and missing namespaces, with RSS namespaced children
		//there is no native way to produce a collection with multiple namespaces
		//getElementsByTagName needs the full tag name, including prefix, so have to use a wildcard
		//cannot use multiple getElementsByTagNameNS requests and merge them, because they will then lose document order
		//could also do this with XPath, and iterate to produce the collection
		const allTags = parentElement.getElementsByTagName('*'), results = [];
		for (let childIndex = 0; childIndex < allTags.length; childIndex++) {
			if (allTags[childIndex].localName === localName && namespaces.indexOf(allTags[childIndex].namespaceURI) !== -1) {
				results[results.length] = allTags[childIndex];
			}
		}
		return results;
	}

	//specific feed manipulation functions

	function getHTMLString (parentElement, parentJSObject) {

		const theType = simplifiedAttr(getAttributeNSs(parentElement, namespaceURI, 'type'), '');
		if (parentElement.children.length) {
			//theType === 'xhtml' || theType.match(/^application\/.*\+xml$/)
			//per spec, xhtml must always begin with a namespaced DIV element containing the XHTML
			//in practice, if there is any child element, it must be XHTML, not HTML or text, since neither of those make sense here
			//detects this DIV if it is either in the XHTML namespace, or not namespaced (where it inherits from the parent)
			if (parentElement.children.length > 1 || parentElement.children[0] !== (parentElement.getElementsByTagNameNS(namespaceXHTML, 'div')[0] || parentElement.getElementsByTagNameNS(namespaceURI[0], 'div')[0]) || hasNonEmptyTextChild(parentElement)) {
				//invalid cases:
				// <content type="xhtml"><span>hello<div>there</div>world</span></content>
				// <content type="xhtml">hello<div>there</div>world</content>
				// <content type="xhtml"><div xmlns="http://www.w3.org/1999/xhtml">hello</div><div xmlns="http://www.w3.org/1999/xhtml">world</div></content>
				// <content type="xhtml"><DIV xmlns="http://www.w3.org/1999/xhtml">hello world</DIV></content>
				//return the element's innerHTML
				return stringifyHTML(parentElement);
			} else {
				//"valid" cases:
				// <content type="xhtml"> <div xmlns="http://www.w3.org/1999/xhtml">hello<div>there</div>world</div> </content>
				// <content type="xhtml"> <div>not<div>really</div>valid</div> </content>
				//return the element's firstchild's innerHTML
				//only preserve the namespace information if the DIV has the XHTML namespace (either prefixed or with xmlns)
				return stringifyHTML(parentElement.children[0], parentElement.children[0].namespaceURI === namespaceXHTML);
			}
		} else if (theType === 'html' || theType === 'text\/html' || theType === 'xhtml' || theType.match(/^application\/.*\+xml$/)) {
			//HTML is supposed to be entity escaped HTML, so that once parsed, the textNode contains HTML
			//however, if XHTML is also specified without any child elements, then it is really HTML, pretending to be XHTML, so XHTML is also accepted here
			//this invalid case will also act like HTML, which is probably the best outcome: <content type="xhtml">hello &amp;lt; world</content>
			return trim(concatenateText(parentElement));
		} else if (!theType || ( theType === 'text' ) || theType.match(/^text\/plain$/) || !theType.match(/^[\-\w]+\/[\-\w\+]+$/)) {
			//catch all text and broken mimetypes (eg. no slash, commas, whitespace)
			return htmlSpecialChars(trimLines(concatenateText(parentElement)));
		} else {
			//base64 encoded attachment, create a link instead, since there is no displayable body
			if (parentJSObject) {
				const dataLink = parentJSObject[parentJSObject.length] = new NewsLink();
				dataLink.url = 'data:'+theType+';base64,'+trim(concatenateText(parentElement)).replace(/\s+/g, '');
				dataLink.mimetype = theType;
			}
			return '';
		}
	}

	function getAutoHTML (parentElement) {
		//in RSS, de-facto behaviour for some elements is that if it looks like HTML, it is probably HTML
		//this is allowed in some versions and not in others (0.91), but reality disagrees - the version is ignored here
		if (parentElement.children.length) {
			//there are child elements, so this must be raw HTML
			return stringifyHTML(parentElement);
		} else {
			//work with the string contents - this even causes CDATA blocks to be handled correctly; any HTML in them will become HTML as needed
			return getAutoHTMLString(concatenateText(parentElement));
		}
	}

	function getImage (imageElement) {
		//images have a URL, and many optional elements that can be used to make a HTML link
		const imageObject = new NewsImage();
		if (imageElement.children.length) {
			//RSS
			for (let childIndex = 0, childElement, childName; childIndex < imageElement.children.length; childIndex++) {
				childElement = imageElement.children[childIndex];
				childName = simplifiedName(childElement.localName);
				if (namespaceURI.indexOf(childElement.namespaceURI) !== -1) {
					if (childName === 'url' || childName === 'link') {
						imageObject[childName] = resolveURL(trim(concatenateText(childElement)), childElement, feedURL);
					} else if (childName === 'feedlink') {
						imageObject.link = resolveURL(trim(concatenateText(childElement)), childElement, feedURL);
					} else if (childName === 'title' || childName === 'feedtitle') {
						imageObject.title = trim(concatenateText(childElement));
					} else if (childName === 'description' || childName === 'width' || childName === 'height') {
						imageObject[childName] = trim(concatenateText(childElement));
					}
				}
			}
		} else {
			//Atom
			imageObject.url = trim(concatenateText(imageElement));
		}
		return imageObject;
	}

	function addLink (linkElement, parentJSObject) {
		//Atom: <link rel="alternate" href="http://example.com/file.html" />
		//Atom: <link rel="enclosure" href="http://example.com/file.mp3" length="123456789" type="audio/mpeg" />
		//RSS: <enclosure url="http://example.com/file.mp3" length="123456789" type="audio/mpeg" />
		//RSS: <link>http://example.com/file.html</link>
		//RSS: <source url="http://example.com/file.html">Some title</source>
		//this blindly gets the Atom rel attribute, even for RSS, which doesn't support it
		//harmless to allow it, so it does
		let rel, linkURL;
		//special RSS cases
		if (linkElement.localName === 'enclosure' || linkElement.localName === 'comments' || linkElement.localName === 'docs') {
			rel = linkElement.localName;
		} else if (linkElement.localName === 'source' || linkElement.localName === 'feedsource') {
			rel = 'via';
		} else {
			//Atom case
			rel = simplifiedAttr(getAttributeNSs(linkElement, namespaceURI, 'rel'), 'alternate');
		}
		if (linkElement.localName !== 'source' && linkElement.localName !== 'feedsource' && linkElement.childNodes.length) {
			//RSS approach, source is handled below instead
			linkURL = trim(concatenateText(linkElement));
		} else if (hasAttributeNSs(linkElement, namespaceURI, 'href')) {
			//Atom approach
			linkURL = trim(getAttributeNSs(linkElement, namespaceURI, 'href'));
		} else if (hasAttributeNSs(linkElement, namespaceURI, 'url')) {
			//RSS approach
			linkURL = trim(getAttributeNSs(linkElement, namespaceURI, 'url'));
		}
		if (linkURL && parentJSObject.links.hasOwnProperty(rel)) {
			//now it will only accept the various types of link that are actually supported
			const newsLink = parentJSObject.links[rel][parentJSObject.links[rel].length] = new NewsLink();
			newsLink.url = resolveURL(linkURL, linkElement, feedURL);
			if (linkElement.localName === 'source' || linkElement.localName === 'feedsource') {
				//RSS is inconsistent even with itself, so this is special cased
				newsLink.title = trim(concatenateText(linkElement));
			} else {
				newsLink.title = trim(getAttributeNSs(linkElement, namespaceURI, 'title'));
			}
			newsLink.mimetype = trim(getAttributeNSs(linkElement, namespaceURI, 'type'));
			newsLink.hreflang = trim(getAttributeNSs(linkElement, namespaceURI, 'hreflang'));
			newsLink.bytelength = trim(getAttributeNSs(linkElement, namespaceURI, 'length'));
		}
	}

	function getCategory (categoryElement) {
		//categories are used for sorting within a view
		const category = new NewsCategory(), categoryContent = trim(concatenateText(categoryElement));
		if (categoryContent.length) {
			//RSS uses a child textNode
			category.label = category.term = categoryContent;
			category.scheme = trim(getAttributeNSs(categoryElement, namespaceURI, 'domain'));
		} else {
			//Atom uses attributes
			category.term = trim(getAttributeNSs(categoryElement, namespaceURI, 'term'));
			category.label = trim(getAttributeNSs(categoryElement, namespaceURI, 'label') || category.term);
			category.scheme = trim(getAttributeNSs(categoryElement, namespaceURI, 'scheme'));
		}
		return category;
	}

	function getGenerator (generatorElement) {
		//the program used to generate the feed - Atom and RSS are compatible, but Atom also has optional attributes
		const generator = new NewsUA();
		generator.name = trim(concatenateText(generatorElement));
		generator.url = trim(getAttributeNSs(generatorElement, namespaceURI, 'uri'));
		generator.version = trim(getAttributeNSs(generatorElement, namespaceURI, 'version'));
		return generator;
	}

	function getPerson (personElement) {
		const person = new NewsPerson();
		if (personElement.children.length) {
			//it is possible for Atom feeds to use native elements or FOAF http://xmlns.com/foaf/spec/ to specify details
			/*
			<feed xmlns="http://www.w3.org/2005/Atom" xmlns:foaf="http://xmlns.com/foaf/0.1" xml:base="http://www.example.org">
				<author>
					<foaf:name>John Doe</foaf:name>
					<foaf:homepage rdf:resource="/~johndoe" />
					<foaf:mbox rdf:resource="mailto:JohnDoe@example.com"/>
				</author>
				<author>
					<name>John Doe</name>
					<email>JohnDoe@example.com</email>
					<uri>http://example.com/~johndoe</uri>
				</author>
			*/
			for (let childIndex = 0, childElement, childName; childIndex < personElement.children.length; childIndex++) {
				childElement = personElement.children[childIndex];
				childName = simplifiedName(childElement.localName);
				if (namespaceURI.indexOf(childElement.namespaceURI) !== -1) {
					//native Atom
					if (childName === 'name' || childName === 'email') {
						person[childName] = trim(concatenateText(childElement));
					} else if (childName === 'uri') {
						person.url = resolveURL(trim(concatenateText(childElement)), childElement, feedURL);
					}
				} else if (childElement.namespaceURI === namespaceFOAF) {
					//FOAF was initially recommended for Atom; https://www.ibm.com/developerworks/library/x-atom10/index.html
					if (childName === 'name') {
						person[childName] = trim(concatenateText(childElement));
					} else if (childName === 'homepage') {
						person.url = resolveURL(trim(childElement.getAttributeNS(namespaceRDF, 'resource')), childElement, feedURL);
					} else if (childName === 'mbox') {
						//FOAF prefixes it with 'mailto:' while native Atom does not
						person.email = trim(childElement.getAttributeNS(namespaceRDF, 'resource')).replace(/^[a-z]+:/i,'');
					}
				}
			}
		} else {
			//RSS only allows an email address
			/*
			<author>JohnDoe@example.com</author>
			*/
			person.email = trim(concatenateText(personElement));
		}
		return person;
	}

	function getCloud (cloudElement) {
		//in practice, this is unlikely to ever be used, even if a feed is found to use it, but it is supported here just in case
		//<cloud domain="example.com" port="80" path="/rpc/path" registerProcedure="NotifyMe" protocol="xml-rpc" />
		const cloud = new NewsCloud();
		cloud.domain = trim(getAttributeNSs(cloudElement, namespaceURI, 'domain'));
		cloud.port = trim(getAttributeNSs(cloudElement, namespaceURI, 'port'));
		cloud.path = trim(getAttributeNSs(cloudElement, namespaceURI, 'path'));
		//attribute will be in lower case when reparsing as HTML due to XML error
		cloud.registerprocedure = trim(getAttributeNSs(cloudElement, namespaceURI, 'registerProcedure')) || trim(getAttributeNSs(cloudElement, namespaceURI, 'registerprocedure'));
		cloud.protocol = trim(getAttributeNSs(cloudElement, namespaceURI, 'protocol'));
		return cloud;
	}

	function getInput (inputElement) {
		//in practice, this is unlikely to ever be used, even if a feed is found to use it, but it is supported here just in case someone wants to create this:
		//<form action="http://example.com/feedback.php" method="GET">Send feedback<input name="textcontent"><input type="submit" value="Submit"></form>
		/*
		<textinput>
			<description>Send feedback</description>
			<title>Submit</title>
			<link>http://example.com/feedback.php?</link>
			<name>textcontent</name>
		</textinput>
		*/
		const input = new NewsInput();
		for (let childIndex = 0, childElement, childName; childIndex < inputElement.children.length; childIndex++) {
			childElement = inputElement.children[childIndex];
			childName = simplifiedName(childElement.localName);
			if (namespaceURI.indexOf(childElement.namespaceURI) !== -1) {
				if (childName === 'description') {
					input.htmldescription = getAutoHTML(childElement);
				} else if (childName === 'title' || childName === 'feedtitle') {
					input.title = trim(concatenateText(childElement));
				} else if (childName === 'name') {
					input.name = trim(concatenateText(childElement));
				} else if (childName === 'link' || childName === 'feedlink') {
					input.link = resolveURL(trim(concatenateText(childElement)), childElement, feedURL);
				}
			}
		}
		return input;
	}

	function addSkipDays (skipElement, parentJSObject) {
		//RSS lists which days to skip when checking the feed - these are referenced by name
		for (let childIndex = 0, childElement, childContent; childIndex < skipElement.children.length; childIndex++) {
			childElement = skipElement.children[childIndex];
			if (namespaceURI.indexOf(childElement.namespaceURI) !== -1 && simplifiedName(childElement.localName) === 'day') {
				childContent = trim(concatenateText(childElement)).toLowerCase();
				if (parentJSObject.hasOwnProperty(childContent)) {
					//only if the day name is listed in the NewsDay object
					parentJSObject[childContent] = true;
				}
			}
		}
	}

	function addSkipHours (skipElement, parentJSObject) {
		//RSS lists which hours to skip when checking the feed - these are referenced by number, in GMT
		for (let childIndex = 0, childElement, childContent; childIndex < skipElement.children.length; childIndex++) {
			childElement = skipElement.children[childIndex];
			if (namespaceURI.indexOf(childElement.namespaceURI) !== -1 && simplifiedName(childElement.localName) === 'hour') {
				childContent = trim(concatenateText(childElement)).toLowerCase();
				if (childContent.match(/^\d+$/) && childContent * 1 < 24 && childContent * 1 >= 0) {
					//only if 0<=number<=23
					parentJSObject[childContent] = true;
				}
			}
		}
	}

	function getItemFeed (itemElement) {
		//get cleaned up representation of the feed (after correcting parsing errors) containing just that single feed item
		//first get the item, then step up the tree and get the markup for each wrapper element in turn
		const xmlSerializer = new XMLSerializer();
		let innerXML = xmlSerializer.serializeToString(itemElement);
		while (itemElement && itemElement.parentNode !== documentNode) {
			itemElement = itemElement.parentNode;
			let parentElement = xmlSerializer.serializeToString(itemElement.cloneNode(false));
			//in XML parsing mode, the element will be self closing <channel /> but in HTML mode it will not <channel></channel>
			//normalise to the non-self-closing version
			parentElement = parentElement.replace(/^<([\w\-:]+)\b([^>]*)\/>$/, '<$1$2></$1>');
			parentElement = parentElement.split('></');
			innerXML = parentElement[0] + '>' + innerXML + '</' + parentElement[1];
		}
		return innerXML;
	}

	function getFeed (channelElement) {
		//get the information about the feed or source element
		const feed = new NewsFeed();
		for (let childIndex = 0, childElement, childName; childIndex < channelElement.children.length; childIndex++) {
			childElement = channelElement.children[childIndex];
			childName = simplifiedName(childElement.localName);
			if (namespaceURI.indexOf(childElement.namespaceURI) !== -1) {
				if (childName === 'image' || childName === 'feedimage') {
					//RSS image object
					if (childElement.children.length) {
						//RSS 0.9 and 1.0 have dummy elements here that point to their counterparts using rdf:resource->rdf:about
						feed.image = getImage(childElement);
					}
				} else if (childName === 'logo') {
					//Atom image object
					feed.image = getImage(childElement);
				} else if (childName === 'icon') {
					//Atom image object for small icons, not a regular image
					feed[childName] = getImage(childElement);
				} else if (childName === 'generator') {
					//Atom only
					feed[childName] = getGenerator(childElement);
				} else if (childName === 'title' || childName === 'feedtitle') {
					//title can be HTML too! Isn't that a fun idea when you want to display it in a mail client?
					if (channelElement.localName === 'channel') {
						//RSS needs automatic detection
						feed.htmltitle = getAutoHTML(childElement);
					} else {
						//Atom is explicit about type
						feed.htmltitle = getHTMLString(childElement);
					}
					feed.plaintitle = getPlainString(feed.htmltitle);
				} else if (childName === 'rights') {
					//Atom lets rights be HTML too
					feed.htmlrights = getHTMLString(childElement);
				} else if (childName === 'subtitle') {
					//subtitle in theory is text only, but realistically, websites might accidentally try to make it (X)HTML
					feed.htmldescription = getHTMLString(childElement);
				} else if (childName === 'copyright') {
					//RSS only - use automatic HTML detection
					feed.htmlrights = getAutoHTML(childElement);
				} else if (childName === 'description') {
					//RSS only - use automatic HTML detection
					feed.htmldescription = getAutoHTML(childElement);
				} else if (childName === 'updated' || childName === 'lastbuilddate' || childName === 'pubdate' || childName === 'rating' || childName === 'language') {
					//simple string
					//updated is Atom, lastbuiddate, pubdate and rating are RSS
					feed[childName] = trim(concatenateText(childElement));
				} else if (childName === 'ttl') {
					//RSS only, numeric string
					feed[childName] = trim(concatenateText(childElement));
				} else if (childName === 'category') {
					//Atom only
					feed[childName][feed[childName].length] = getCategory(childElement);
				} else if (childName === 'link' || childName === 'feedlink' || childName === 'docs') {
					//link is RSS and Atom, docs is RSS
					addLink(childElement, feed);
				} else if (childName === 'cloud') {
					//RSS
					feed[childName][feed[childName].length] = getCloud(childElement);
				} else if (childName === 'author' || childName === 'contributor' || childName === 'managingeditor' || childName === 'webmaster') {
					//author and contributor are Atom only, managingeditor and webmaster are RSS only
					feed.people[childName][feed.people[childName].length] = getPerson(childElement);
				} else if (childName === 'textinput') {
					//RSS
					if (childElement.children.length) {
						//RSS 0.9 and 1.0 have dummy elements here that point to their counterparts using rdf:resource->rdf:about
						feed[childName][feed[childName].length] = getInput(childElement);
					}
				} else if (childName === 'skipdays') {
					//RSS
					addSkipDays(childElement, feed[childName]);
				} else if (childName === 'skiphours') {
					//RSS
					addSkipHours(childElement, feed[childName]);
				}
			} else if (childElement.namespaceURI === namespaceSyndication) {
				//handling for polling frequency
				//<sy:updatePeriod>daily</sy:updatePeriod>
				//<sy:updateFrequency>1</sy:updateFrequency>
				if (childName === 'updateperiod' || childName === 'updatefrequency') {
					feed[childName] = trim(concatenateText(childElement)).toLowerCase();
				}
			//} else if (childElement.namespaceURI === OTHER_URI) {
				//handling for elements from other namespaces can go here
			}
		}
		if (!feed.language && documentNode.documentElement.hasAttribute('xml:lang')) {
			//strictly speaking, this could be on any child element, but for the purpose of a whole Atom feed, it is only recognised on the root element
			feed.language = trim(documentNode.documentElement.getAttribute('xml:lang'));
		}
		//combine properties from RSS and Atom to make useful date management properties
		//in practice, pubdate and lastbuilddate are identical, since RSS feeds only supports a single date per item
		//however, pubdate should be favoured for publish date and lastbuilddate should be favoured for last-updated
		feed.publishdate = feed.pubdate || feed.updated || feed.lastbuilddate;
		feed.updatedate = feed.lastbuilddate || feed.updated || feed.pubdate;
		//try to constuct a reliable ttl
		if (feed.ttl.match(/^\d+$/)) {
			//get an integer, or reset to default for the invalid value 0
			feed.ttl = parseInt(feed.ttl) || '';
		} else if (feed.updateperiod || feed.updatefrequency) {
			//construct a ttl from updateperiod and updatefrequency
			let timeFrequency;
			if (feed.updatefrequency.match(/^\d+$/)) {
				//default in the spec is 1, which will also be used here if the invalid value 0 is used
				timeFrequency = parseInt(feed.updatefrequency) || 1;
			} else {
				timeFrequency = 1;
			}
			const timeMultipliers = { hourly: 60, daily: 24 * 60, weekly: 7 * 24 * 60, monthly: 7 * 24 * 60 * 365 / 12, yearly: 7 * 24 * 60 * 365 };
			if (timeMultipliers.hasOwnProperty(feed.updateperiod)) {
				//number of minutes
				feed.ttl = Math.round(timeMultipliers[feed.updateperiod] / timeFrequency);
			} else {
				//default is daily
				feed.ttl = Math.round(timeMultipliers.daily / timeFrequency);
			}
		} else {
			//reset to default
			feed.ttl = '';
		}
		return feed;
	}

	//parsing begins

	//get the important wrapper elements that the feed needs to contain

	let hasErrors = false;
	if (typeof(htmlString) !== 'string') {
		return 'HTMLStringNotString';
	}
	if (typeof(feedURL) !== 'string') {
		return 'FeedURLNotString';
	}
	if (!documentNode || documentNode.getElementsByTagNameNS(namespaceXHTML, 'parsererror').length) {
		//if it starts with an XML opening tag, and has at least 4 opening and 4 closing tags, it is likely to be XML rather than binary data
		if (htmlString && htmlString.match(/^\w*</) && htmlString.match(/<[\w\-:]+[^>]*>/g).length > 3 && htmlString.match(/<\/[\w\-:]+>/g).length > 3) {
			//reparse the XML string as HTML, and hope it retains some meaning that can be extracted
			//RSS/Atom elements get mistaken for self-closing/special HTML equivalents, which causes invalid nesting in HTML mode, so replace with alternative names which are handled below
			//try to be nice with legitimate <source> elements inside a media element, by non-greedy matching when there is a closing tag as well
			//potentially this could mis-match when a HTML string inside an Atom <source> contains an HTML <source> element, but it is highly unlikely in that location
			//title is magical in HTML and has all nested tags converted to entities, so that is also covered here
			/*
			deep breath - this regex is based on https://docstore.mik.ua/orelly/perl/cookbook/ch06_16.htm
			/
			 <                         tag start
			 (source|link|image)       with one of these tag names + capture
			 \b                        end of the name
			 (                         capture
			  ([^\/>]|\/[^>])*         not self-closing XML - forward slash cannot appear immediately before >
			  >                        end of tag
			  [^<]*                    any non tag-starts
			  (                        repeating group
			   (?!                     negative lookahead - must not be the start of
			    <\1\b                  another opening tag with the same name as before
			    ([^\/>]|\/[^>])*       which is also not self-closing XML
			    >                      end of the opening tag
			   )                       end of lookahead
			   <                       start of a tag which did not match the lookahead
			   [^<]*                   any non tag-starts
			  )*?                      non-greedy any number of times
			 )                         end of capture
			 <\/\1>                    closing tag matching the first tag
			/ig                        case insensitive+global
			*/
			//alert(
			//	'a<source>b<source/>c<source>d<source>e<source>only this parentNode<source />should change</source>f</source>g<source   />h<source>i<source>j'
			//	.replace(/<(source|link|image|title)\b(([^\/>]|\/[^>])*>[^<]*((?!<\1\b([^\/>]|\/[^>])*>)<[^<]*)*?)<\/\1>/ig,'<feed$1$2</feed$1>')
			//);
			while (htmlString.match(/<(source|link|image|title)\b(([^\/>]|\/[^>])*>[^<]*((?!<\1\b([^\/>]|\/[^>])*>)<[^<]*)*?)<\/\1>/i)) {
				//these elements can be nested within each other, and it will fail to match the inner one in such cases, so it needs to be done in multiple passes
				htmlString = htmlString.replace(/<(source|link|image|title)\b(([^\/>]|\/[^>])*>[^<]*((?!<\1\b([^\/>]|\/[^>])*>)<[^<]*)*?)<\/\1>/ig,'<feed$1$2</feed$1>');
			}
			//self closing XML tags like <foo:link /> and <category /> cause invalid nesting in HTML mode, so try to close them
			//unprefixed <link> does not need special handling, since it is self closing in both XML and HTML
			//theoretically, this could list all RSS/Atom tags instead of just the ones that are expected to use the self-closing syntax, but the more
			//changes it makes, the more it could accidentally mess up ... <foo bar=">" /> ... keep it simple to avoid breakage
			htmlString = htmlString.replace(/<(\w+:link|\w+:category|category|\w+:cloud|cloud|\w+:enclosure|enclosure)\b([^>]*)\/>/ig, '<$1$2></$1>');
			//convert CDATA into escaped HTML
			//errors are most likely to happen with RSS feeds that use malformed HTML in the description without CDATA, so this probably will not happen often
			htmlString = htmlString.replace(/<!\[CDATA\[([\w\W]*?)\]\]>/ig, function (match, capture) {
				return htmlSpecialChars(capture);
			});
			htmlString = '<!doctype html><html><head><title></title></head><body>' + htmlString + '</body></html>';
			//this line causes a massive hangup when loading an image instead of a proper document, so it is protected using RegEx to scan for likely XML tags
			//binary data may also contain what looks like markup, but it is very unlikely to happen by chance
			documentNode = (new DOMParser()).parseFromString(htmlString, 'text/html');
			hasErrors = true;
		}
		if (!documentNode || documentNode.getElementsByTagNameNS(namespaceXHTML, 'parsererror').length) {
			//in theory, DOMParser should never create a parse error, since it is using the HTML parser, but binary data can still land here
			return 'NoDocumentNode';
		}
		if (documentNode.body) {
			//after loading it as HTML, there is a HTML document wrapped around it (forced by the code above, in case the parser fails to create it)
			documentNode = documentNode.body;
			documentNode.documentElement = documentNode.children[0];
		}
	}
	if (!documentNode.documentElement) {
		return 'NoDocumentElement';
	}
	const namespaceURI = [documentNode.documentElement.namespaceURI];
	if (!namespaceURI[0] && documentNode.documentElement.localName === 'feed') {
		//namespace was not used on an Atom feed, which is invalid but supported - also need to keep checking for the default namespace
		//no point in doing this for RSS because RSS 0.91, 0.92 and 2.0 do not use a default namespace
		//evil case, which is supported just in case a malformed feed matches this pattern
		// <feed>
		// <item xmlns="http://www.w3.org/2005/Atom" />
		// <item />
		namespaceURI[1] = namespaceAtom;
	} else if (namespaceURI[0] === namespaceRDF) {
		//RSS 1.0 uses the RDF namespace for the root node - everything else of importance gets the RSS 1.0 namespace
		//RSS 0.9 also uses the RDF namespace for the root node - everything else of importance gets the RSS 0.9 namespace
		namespaceURI[1] = namespaceRSS;
		namespaceURI[2] = namespaceRSS09;
	}
	let channelElement;
	if (documentNode.documentElement.localName === 'feed') {
		//Atom has the feed details directly inside the documentElement
		channelElement = documentNode.documentElement;
	} else {
		//RSS has a separate channel element
		channelElement = getElementsByTagNameNSs(documentNode.documentElement, namespaceURI, 'channel')[0];
	}
	if (!channelElement) {
		return 'NoChannelElement';
	}

	//parse the metadata from the feed iteself
	const feed = getFeed(channelElement);
	feed.haserrors = hasErrors;
	//RSS 0.9 and 1.0 put image and textInput as siblings of the channel element, while 0.91, 0.92 and 2.0 put them inside the channel element
	//rather than checking for the RDF root node, it just checks if the channel element has a lot of siblings, since this will catch more
	//error cases, such as a feed using the RSS 1.0 image-in-root-node, but the RSS 2.0 tag names
	if (documentNode.documentElement !== channelElement && documentNode.documentElement.children.length > 1) {
		//RSS always has the separate channelElement, but 0.91, 0.92 and 2.0 will only have a single child; the channel element
		for (let childIndex = 0, childElement, childName; childIndex < documentNode.documentElement.children.length; childIndex++) {
			childElement = documentNode.documentElement.children[childIndex];
			childName = simplifiedName(childElement.localName);
			if (namespaceURI.indexOf(childElement.namespaceURI) !== -1) {
				//like items, these two elements can be siblings of the channel element - referred to by rdf:resource->rdf:about in 1.0
				if (childName === 'image' || childName === 'feedimage') {
					feed.image = getImage(childElement);
				} else if (childName === 'textinput') {
					feed[childName][feed[childName].length] = getInput(childElement);
				}
			}
		}
	}

	//parse each news item
	//RSS 0.9 and 1.0 put items as siblings of the channel element, while 0.91, 0.92 and 2.0 put them inside the channel element
	//therefore this code always looks for them anywhere inside the document
	let allItems;
	if (channelElement.localName === 'feed') {
		allItems = getElementsByTagNameNSs(documentNode, namespaceURI, 'entry');
	} else {
		allItems = getElementsByTagNameNSs(documentNode, namespaceURI, 'item');
	}
	for (let itemIndex = 0; itemIndex < allItems.length; itemIndex++) {
		feed.feeditems[itemIndex] = new NewsItem();
		//each item contains a series of elements whose contents provide the (meta)data for that item
		for (let itemChildIndex = 0, itemChild, itemChildName; itemChildIndex < allItems[itemIndex].children.length; itemChildIndex++) {
			itemChild = allItems[itemIndex].children[itemChildIndex];
			itemChildName = simplifiedName(itemChild.localName);
			if (namespaceURI.indexOf(itemChild.namespaceURI) !== -1) {
				if (itemChildName === 'link' || itemChildName === 'feedlink' || itemChildName === 'enclosure' || itemChildName === 'comments') {
					//link is Atom, enclosure is RSS
					addLink(itemChild, feed.feeditems[itemIndex]);
				} else if (itemChildName === 'title' || itemChildName === 'feedtitle') {
					//title can be HTML here too! More fun for a mail client.
					if (itemChild.parentNode.localName === 'item') {
						//RSS needs automatic detection
						feed.feeditems[itemIndex].htmltitle = getAutoHTML(itemChild);
					} else {
						//Atom is explicit about type
						feed.feeditems[itemIndex].htmltitle = getHTMLString(itemChild);
					}
					feed.feeditems[itemIndex].plaintitle = getPlainString(feed.feeditems[itemIndex].htmltitle);
				} else if (itemChildName === 'summary' || itemChildName === 'rights') {
					//summary and rights are Atom only
					feed.feeditems[itemIndex]['html'+itemChildName] = getHTMLString(itemChild);
				} else if (itemChildName === 'content') {
					//Atom, same as above, but can sometimes provide mimetypes that need to become an enclosure link instead
					feed.feeditems[itemIndex].htmlcontent = getHTMLString(itemChild, feed.feeditems[itemIndex].links.enclosure);
				} else if (itemChildName === 'description') {
					//RSS only - use automatic HTML detection
					feed.feeditems[itemIndex].htmldescription = getAutoHTML(itemChild);
				} else if (itemChildName === 'author' || itemChildName === 'contributor') {
					//author is Atom and RSS, contributor is Atom only
					feed.feeditems[itemIndex].people[itemChildName][feed.feeditems[itemIndex].people[itemChildName].length] = getPerson(itemChild);
				} else if (itemChildName === 'category') {
					//Atom only
					feed.feeditems[itemIndex][itemChildName][feed.feeditems[itemIndex][itemChildName].length] = getCategory(itemChild);
				} else if (itemChildName === 'id' || itemChildName === 'published' || itemChildName === 'updated' || itemChildName === 'pubdate') {
					//Atom's ID is intentionally *not* unique since an item may update and both an old and new version can be in the same feed(!)
					//however, they are supposed to refer to the same article - normally, the latest one would be used, but this implies
					//that the dates are valid, regognised, and different - reality may vary
					//Atom:
					// <id>http://example.com/1234</id>
					// <published>2006-06-09T18:18:53Z</published>
					// <updated>2006-06-16T18:10:35Z</updated>
					//RSS:
					// <pubDate>Thu, 5 Apr 2012 10:35:00 +0000</pubDate>
					feed.feeditems[itemIndex][itemChildName] = trim(concatenateText(itemChild));
				} else if (itemChildName === 'guid') {
					//guid is the RSS equivalent of Atom's id, except it is supposed to be unique (reality may vary)
					//guid by default is also a link to the article
					feed.feeditems[itemIndex].id = trim(concatenateText(itemChild));
					//attribute will be in lower case when reparsing as HTML due to XML error
					if (simplifiedAttr(getAttributeNSs(itemChild, namespaceURI, 'isPermaLink'), '') !== 'false' && simplifiedAttr(getAttributeNSs(itemChild, namespaceURI, 'ispermalink'), '') !== 'false') {
						//create a link as well
						addLink(itemChild, feed.feeditems[itemIndex]);
					}
				} else if (itemChildName === 'source' || itemChildName === 'feedsource') {
					//two completely different elements with the same name, serving completely different purposes
					if (itemChild.parentNode.localName === 'item') {
						//RSS this is just a link
						addLink(itemChild, feed.feeditems[itemIndex]);
					} else {
						//Atom, this element can contain any meta data that a feed can contain
						feed.feeditems[itemIndex].source = getFeed(itemChild);
					}
				}
			//} else if (childElement.namespaceURI === OTHER_URI) {
				//handling for elements from other namespaces can go here
			}
		}
		//combine properties from RSS and Atom to make useful date management properties
		//Atom properties make sense here, but authors may omit one while supplying the other, so check for both, just in case
		feed.feeditems[itemIndex].publishdate = feed.feeditems[itemIndex].published || feed.feeditems[itemIndex].pubdate || feed.feeditems[itemIndex].updated;
		feed.feeditems[itemIndex].updatedate = feed.feeditems[itemIndex].updated || feed.feeditems[itemIndex].pubdate || feed.feeditems[itemIndex].published;
		//Atom allows authors to specify a summary (like a first paragraph), or content (the whole body), or both
		//the body is: if possible the Atom content, or simply the RSS description, or if all else fails then the Atom summary
		feed.feeditems[itemIndex].htmlbody = feed.feeditems[itemIndex].htmlcontent || feed.feeditems[itemIndex].htmldescription || feed.feeditems[itemIndex].htmlsummary;
		feed.feeditems[itemIndex].xmloriginal = getItemFeed(allItems[itemIndex]);
	}

	return feed;

};

FeedParser.prototype.parseOPML = function (documentNode, feedURL) {

	function addOutlines (parentElement, parentJSObject) {
		//recursive function creates the folder structure
		for (let childIndex = 0, childElement, childName; childIndex < parentElement.children.length; childIndex++) {
			childElement = parentElement.children[childIndex];
			childName = simplifiedName(childElement.localName);
			if (namespaceURI.indexOf(childElement.namespaceURI) !== -1 && childName === 'outline' && simplifiedAttr(getAttributeNSs(childElement, namespaceURI, 'isComment'), '') !== 'true') {
				//isComment is poorly specified (like most of OPML), but appears to make the <outline> and its children into an XML comment - therefore ignored
				const oneOutline = new NewsOutline();
				//OPML attributes are not case sensitive
				if (['rss', 'atom'].indexOf(simplifiedAttr(getAttributeNSs(childElement, namespaceURI, 'type'), '')) !== -1) {
					//officially, type="rss" represents all newsfeeds, not just RSS, but "atom" is also allowed here because some feed readers mistakenly export it
					oneOutline.feedurl = trim(getAttributeNSs(childElement, namespaceURI, 'xmlUrl'));
					if (oneOutline.feedurl) {
						//the only other essential attribute, it can be added to the collection
						oneOutline.feedurl = resolveURL(oneOutline.feedurl, childElement, feedURL);
						oneOutline.created = trim(getAttributeNSs(childElement, namespaceURI, 'created'));
						//the spec does not say if the description can be HTML, but both RSS and Atom allow it to have HTML
						//default to a blank string, getAutoHTMLString does not cast
						oneOutline.htmldescription = getAutoHTMLString(getAttributeNSs(childElement, namespaceURI, 'description') || '');
						oneOutline.language = trim(getAttributeNSs(childElement, namespaceURI, 'language'));
						//page URL is supposed to htmlUrl per spec, but some feed readers export url instead
						oneOutline.pageurl = trim(getAttributeNSs(childElement, namespaceURI, 'htmlUrl')) || trim(getAttributeNSs(childElement, namespaceURI, 'url'));
						if (oneOutline.pageurl) {
							oneOutline.pageurl = resolveURL(oneOutline.pageurl, childElement, feedURL);
						}
						//version is an optional attribute with no official values, that says what type of feed it is
						//RSS normally means RSS 0.91, 0.92 and 2.0
						//RSS1 normally means RSS 0.9 and 1.0
						//scriptingNews means scriptingNews
						//there is no official value for Atom, but feeds and examples have been found with "Atom" and "Atom 1.0"
						//because it is not standardised, it is recognised as simply "likely to be valid/invalid"
						//in addition, it could be pointing to an RSS feed, but say "Atom", since a feed reader creates it when it initially saves the feed
						//and it may change format later - therefore allow all formats to be imported, and simply flag "may not be supported"
						const supported = simplifiedAttr(getAttributeNSs(childElement, namespaceURI, 'version'), '');
						oneOutline.supported = (!supported || supported.match(/^(rss\d*|atom\s*[\d\.]*)$/i));
					}
				}
				//categories can either be explicit, or specified by a parent folder - copy them here onto the child for easy access
				if (hasAttributeNSs(childElement, namespaceURI, 'category')) {
					oneOutline.category = trim(getAttributeNSs(childElement, namespaceURI, 'category'));
				} else {
					oneOutline.category = parentJSObject.category;
				}
				if (childElement.children.length) {
					//check if this outline is being used as a folder
					addOutlines(childElement, oneOutline);
				}
				if (oneOutline.feedurl || oneOutline.items.length) {
					//it is either a feed, or a folder containing a feed, add it
					parentJSObject.items[parentJSObject.items.length] = oneOutline;
					//per spec, title should be omited, and text is the name of a folder or a feed, but some feed readers export title instead
					//per spec, text attributes MAY contain encoded HTML markup
					oneOutline.htmltitle = getAutoHTMLString(trim(getAttributeNSs(childElement, namespaceURI, 'text')) || trim(getAttributeNSs(childElement, namespaceURI, 'title')));
					oneOutline.plaintitle = getPlainString(oneOutline.htmltitle);
				}
			}
		}
	}

	//OPML documents must be well formatted XML
	//root element is 'opml' but before v1, it was 'outlineDocument'
	//before v2, there was no namespace (null) - as of v2, the namespace is only supposed to be used when embedding in other XML formats, not inside an OMPL document
	//it is allowed here, just in case
	//default is null but empty strings are allowed, so that xmlns="" can work
	const namespaceURI = [null, '','http://opml.org/spec2'];
	if (typeof(feedURL) !== 'string') {
		return 'FeedURLNotString';
	}
	if (!documentNode || documentNode.getElementsByTagNameNS(namespaceXHTML, 'parsererror').length) {
		return 'NoDocumentNode';
	}
	if (!documentNode.documentElement) {
		return 'NoDocumentElement';
	}
	if ((documentNode.documentElement.localName !== 'opml' && documentNode.documentElement.localName !== 'outlineDocument') || namespaceURI.indexOf(documentNode.documentElement.namespaceURI) === -1) {
		return 'WrongDocumentElement';
	}
	const opml = new NewsList();
	for (let childIndex = 0, childElement, childName; childIndex < documentNode.documentElement.children.length; childIndex++) {
		childElement = documentNode.documentElement.children[childIndex];
		if (namespaceURI.indexOf(documentNode.documentElement.namespaceURI) !== -1) {
			childName = simplifiedName(childElement.localName);
			//OPML documents are supposed to contain a head and a body
			if (childName === 'head') {
				//a series of random elements, with text nodes containing the values - a few element names are semi-standard, and can be used here
				for (let headIndex = 0, headChild, headName; headIndex < childElement.children.length; headIndex++) {
					headChild = childElement.children[headIndex];
					if (namespaceURI.indexOf(headChild.namespaceURI) !== -1) {
						headName = simplifiedName(headChild.localName);
						if (headName === 'title') {
							opml.htmltitle = getAutoHTMLString(concatenateText(headChild));
							opml.plaintitle = getPlainString(opml.htmltitle);
						} else if (headName === 'ownername' || headName === 'owneremail') {
							opml[headName] = trim(concatenateText(headChild));
						} else if (headName === 'datecreated') {
							opml.publishdate = trim(concatenateText(headChild));
						} else if (headName === 'datemodified') {
							opml.updatedate = trim(concatenateText(headChild));
						} else if (headName === 'ownerid') {
							opml.ownerurl = resolveURL(trim(concatenateText(headChild)), headChild, feedURL);
						}
					}
				}
			} else if (childName === 'body') {
				//a series of potentially nested <outline> elements
				addOutlines(childElement, opml);
			}
		}
	}
	return opml;
};
